using System;
using System.Iec61131Lib;
using Eclr;
using RTC_APILib;
using RTC_CARD_Lib;

namespace CardReader
{
    [FUNCTION_BLOCK]
    unsafe public class CardReader
    {
        // 萔`
        //------------------------------------------------------------------------------
        // RS232-C֘A
        //------------------------------------------------------------------------------
        // IO Port Address
        private const UInt16 IO_BASE = 0x3f8;           // IO|[gx[XAhX
        private const UInt16 IO_INTMSK = IO_BASE + 1;   // 荞݃}XNWX^
        private const UInt16 IO_INTRR = IO_BASE + 2;    // 荞ݗvWX^
        private const UInt16 IO_LCTL = IO_BASE + 3;     // CRg[WX^
        private const UInt16 IO_MCTL = IO_BASE + 4;     // fRg[WX^
        private const UInt16 IO_LSTS = IO_BASE + 5;     // CXe[^XWX^
        private const UInt16 IO_MSTS = IO_BASE + 6;     // fXe[^XWX^

        // 荞ݗvWX^Bit`
        private const Byte IIR_PENDING = 0x01;          // 荞ݗL

        // fXe[^XWX^Bit`
        private const Byte MST_CTSx = 0x01;             // (CTS)MM̕ω
        private const Byte MST_DSRx = 0x02;             // (DSR)f[^ZbgfBM̕ω
        private const Byte MST_DCDx = 0x08;             // (DCD)LAoM̕ω
        private const Byte MST_CTS = 0x10;              // (CTS)MM
        private const Byte MST_DSR = 0x20;              // (DSR)f[^ZbgfBM
        private const Byte MST_DCD = 0x80;              // (DCD)LAoM

        // J[h[_ɑ΂Mt`
        private const Byte RDDx = MST_CTSx;             // (RDD)J[hf[^M̕ω
        private const Byte RCPx = MST_DSRx;             // (RCP)TvOM̕ω
        private const Byte CLSx = MST_DCDx;             // (CLS)J[hsM̕ω
        private const Byte RDD = MST_CTS;               // (RDD)J[hf[^M
        private const Byte RCP = MST_DSR;               // (RCP)TvOM
        private const Byte CLS = MST_DCD;               // (CLS)J[hsM

        //------------------------------------------------------------------------------
        // 荞݃x`
        //------------------------------------------------------------------------------
        private const UInt16 IRQ4_LEVEL = 0x48;


        // ^XNϐ
        private UInt32 RDD_Cnt;                         // J[hf[^i[
        private RtMemory cmemData = null;               // J[hۑ惁
        private UInt32* pReadCnt = null;                // J[hǂݎ񐔂̕ۑAhX
        private Byte* pbData = null;                    // J[hf[^ۑ̐擪AhX
        private Int32 Err = 0;                          // G[


        // *** RT-CvO֐Q ******************************************
        /// <summary>
        /// RXgN^
        /// </summary>
        public CardReader()
        {
        }

        /// <summary>
        /// fXgN^
        /// </summary>
        ~CardReader()
        {
        }

        /// <summary>
        /// 
        /// </summary>
        public void __Init()
        {
            Init_Memory();          // J[hf[^ۑp̏

            Init_Interrupt();       // 荞݃Xbh̏

            Init_Device();          // COM|[g̏
        }

        /// <summary>
        /// s
        /// </summary>
        public void __Process()
        {
            Byte byState;     // fXe[^XWX^

            //-----------------------------------------------------------------
            // 荞݃nh̃Cxgʒmɂ{vZX쓮܂
            // J[h[_̊eM(CLS,RCP,RDD)͕_̂߁A
            // t(High  Low)ɂȂĂ邱ƂɒӂĂ
            //-----------------------------------------------------------------

            if (Err != 0) return;   // G[Ăꍇ́AI

            //-----------------------------------------------------------------
            // 荞ݗṽ`FbN (荞ݗvȂȂ܂ŏ[v)
            //-----------------------------------------------------------------
            while ((pcibus.InBYTE(IO_INTRR) & IIR_PENDING) == 0x00)
            {
                //---------------------------------------------------------
                // CLS(J[hsM)LoẘԁAf[^ǂݍݏ[v
                //---------------------------------------------------------
                byState = 0x00;   // ͕K[vɓ邽߁A0x00Zbg
                while ((byState & CLS) == 0x00)
                {
                    // fXe[^XWX^Ǎ
                    byState = pcibus.InBYTE(IO_MSTS);

                    // fXe[^XWX^`FbNAf[^i[
                    Set_CardData(byState);
                }
            }
        }
        // *** ֐Q ****************************************************
        /// <summary>
        /// J[hf[^i[
        /// </summary>
        /// <param name="State"></param>
        private void Set_CardData(Byte State)
        {
            //--------------------------------------------------------------//
            // CLS(J[hsM)`FbN                                  //
            //--------------------------------------------------------------//
            // CLSMԕωm
            if ((State & CLSx) == CLSx)
            {
                // J[hsJn(CLSM:Low)̏ꍇ
                if ((State & CLS) == 0x00)
                {
                    RDD_Cnt = 0;                // f[^i[񐔂NA

                    for (int i = 0; i < 100; i++)
                    {
                        pbData[i] = 0;          // f[^ۑ惁̃NA
                    }
                }
                // J[hsI(CLSM:High)̏ꍇ
                else
                {
                    (*pReadCnt)++;              // ǂݎ񐔂CNg
                }
            }

            //--------------------------------------------------------------//
            // RCP(TvOM)`FbN                        //
            // RCP̗̃^C~ORDD(J[hf[^M)ǂݍ     //
            //--------------------------------------------------------------//
            // RCP̗m(RCPωARCP:Low)
            if ((State & (RCP | RCPx)) == RCPx)
            {
                //----------------------------------------------------------//
                // RDD(J[hf[^)1biti[                      //
                //----------------------------------------------------------//
                // RDDM:Low̏ꍇ
                if ((State & RDD) == 0x00)
                {
                    // f[^rbg[1]ɐݒ
                    pbData[RDD_Cnt / 8] |= (Byte)(0x01 << (int)(RDD_Cnt % 8));
                }
                // RDDM:High̏ꍇ
                else
                {
                    // f[^rbg[0]ɐݒ
                    pbData[RDD_Cnt / 8] &= (Byte)~(0x01 << (int)(RDD_Cnt % 8));
                }

                //----------------------------------------------------------//
                //  f[^i[񐔂CNg                          //
                //  Note: J[h̐擪0xFFł邽߁AJEgJn    //
                //        RDDLow̏ꍇƂ                        //
                //----------------------------------------------------------//
                if (((RDD_Cnt == 0) && ((State & RDD) == 0x00))
                    || (RDD_Cnt > 0))
                {
                    RDD_Cnt++;
                }
            }
        }

        /// <summary>
        /// L֘A
        /// </summary>
        private void Init_Memory()
        {
            void* pMem;

            IecString80 MEM_NAME = new IecString80();
            MEM_NAME.ctor();
            MEM_NAME.s.Init("CardReader");      // L̃J^O

            // J[hf[^ۑ惁m
            cmemData = new RtMemory(4096, (Byte)RTC_APILib.HandleSelection.ROOT_PROCESS, MEM_NAME);
            if (cmemData.GetLastErr() != 0)
            {
                // łɃ̈悪ꍇ͎QƂ
                IecString80 sProcName = new IecString80();
                sProcName.ctor();
                sProcName.s.Init("");

                // w薼"CardReader"̃̈QƂ܂
                cmemData = new RtMemory(sProcName, MEM_NAME);
                if (cmemData.GetLastErr() != 0)
                {
                    Err = 1;    // QƃG[
                    RT.printf("RtMemory Refer Error\n");
                    return;
                }
            }
            // ̈̐擪AhX擾
            pMem = cmemData.GetMemPtr();
            if (pMem == null)
            {
                RT.printf("memGet Error\n");
                Err = 1;    // mۃG[
                return;
            }

            pReadCnt = (UInt32*)pMem;       // ̐擪4byteɃJ[hǂݎ񐔂i[
            pbData = (Byte*)&pReadCnt[1];   // 5byteڈȍ~ɃJ[hf[^i[
            *pReadCnt = 0;                  // J[hǂݎ񐔂
        }

        /// <summary>
        /// 荞݃Xbh
        /// </summary>
        private void Init_Interrupt()
        {
            UInt16 hINTRR;

            // 荞݃nhEXbh̃N[Y(s͖܂)
            CARD.RTC_CloseInterruptThread(IRQ4_LEVEL);

            // 荞݃nhEXbh
            hINTRR = CARD.RTC_CreateInterruptThread(0, IRQ4_LEVEL);   // Cxgԍ:0, 荞݃x:IRQ4
        }

        /// <summary>
        /// foCX
        /// </summary>
        private void Init_Device()
        {
            // Com|[g̏s܂
            pcibus.OutBYTE(IO_MCTL, 0x00);              // fRg[: ER RS off /荞ݖ
            Delay();
            pcibus.OutBYTE(IO_LCTL, 0x80);              // DLAB=1 ({[[gݒ胂[hON)
            Delay();
            pcibus.OutHWORD(IO_BASE, 12);               // {[[g: 12 = 9600 Kbps
            Delay();
            pcibus.OutBYTE(IO_LCTL, 0x03);              // ʐMݒ:peB=, Xgbvrbg=1bit, f[^=8bit  ({[[gݒ胂[hOFF)
            Delay();
            pcibus.OutBYTE(IO_INTRR, 0x06);             // FIFO: gp
            Delay();
            pcibus.OutBYTE(IO_MCTL, 0x0b);              // fRg[: ER RS ON / 荞ݗL
            Delay();
            pcibus.OutBYTE(IO_INTMSK, 0x08);            // 荞݃}XN: fXe[^X荞ݗL
            Delay();

            //-----------------------------------------------------------------
            // 荞ݗvNA
            //-----------------------------------------------------------------
            // 荞ݗvȂȂ܂ŏ[v
            while ((pcibus.InBYTE(IO_INTRR) & 0x01) == 0x00)
            {
                // fXe[^XWX^Readɂ芄ݗvNA܂
                pcibus.InBYTE(IO_MSTS);
            }
        }

        /// <summary>
        /// x(foCXւ̐ݒ̐}CNbҋ@)
        /// </summary>
        private void Delay()
        {
            pcibus.InBYTE(0x80);
        }
    }
}
